package org.xbib.datastructures.immutable;

import java.util.ListIterator;
import java.util.NoSuchElementException;

/**
 * This class provides a skeletal implementation of the {@link ListIterator} interface across a
 * fixed number of elements that may be retrieved by position. It does not support {@link #remove},
 * {@link #set}, or {@link #add}.
 */
abstract class AbstractIndexedListIterator<E extends Object>
        extends UnmodifiableListIterator<E> {
    private final int size;
    private int position;

    /**
     * Returns the element with the specified index. This method is called by {@link #next()}.
     */
    protected abstract E get(int index);

    /**
     * Constructs an iterator across a sequence of the given size whose initial position is 0. That
     * is, the first call to {@link #next()} will return the first element (or throw {@link
     * NoSuchElementException} if {@code size} is zero).
     *
     * @throws IllegalArgumentException if {@code size} is negative
     */
    protected AbstractIndexedListIterator(int size) {
        this(size, 0);
    }

    /**
     * Constructs an iterator across a sequence of the given size with the given initial position.
     * That is, the first call to {@link #nextIndex()} will return {@code position}, and the first
     * call to {@link #next()} will return the element at that index, if available. Calls to {@link
     * #previous()} can retrieve the preceding {@code position} elements.
     *
     * @throws IndexOutOfBoundsException if {@code position} is negative or is greater than {@code
     *                                   size}
     * @throws IllegalArgumentException  if {@code size} is negative
     */
    protected AbstractIndexedListIterator(int size, int position) {
        checkPositionIndex(position, size, "index");
        this.size = size;
        this.position = position;
    }

    @Override
    public final boolean hasNext() {
        return position < size;
    }

    @Override
    public final E next() {
        if (!hasNext()) {
            throw new NoSuchElementException();
        }
        return get(position++);
    }

    @Override
    public final int nextIndex() {
        return position;
    }

    @Override
    public final boolean hasPrevious() {
        return position > 0;
    }

    @Override
    public final E previous() {
        if (!hasPrevious()) {
            throw new NoSuchElementException();
        }
        return get(--position);
    }

    @Override
    public final int previousIndex() {
        return position - 1;
    }

    private static int checkPositionIndex(int index, int size, String desc) {
        if (index < 0 || index > size) {
            throw new IndexOutOfBoundsException(badPositionIndex(index, size, desc));
        }
        return index;
    }

    private static String badPositionIndex(int index, int size, String desc) {
        if (index < 0) {
            return String.format("%s (%s) must not be negative", desc, index);
        } else if (size < 0) {
            throw new IllegalArgumentException("negative size: " + size);
        } else { // index > size
            return String.format("%s (%s) must not be greater than size (%s)", desc, index, size);
        }
    }
}
